mod arithmetic {
    use std::rc::Rc;

    use crate::{environment, node::{
        add::Add, constant::Constant, divide::Divide, multiply::Multiply, subtract::Subtract, Node, NodeEnum
    }};
    use environment::Environment;

    use test_case::test_case;

    #[test_case(69.0, 420.0, 489.0 ; "when both are positive")]
    #[test_case(-2.0,  -4.0, -6.0  ; "when both are negative")]
    #[test_case(0.0,  0.0, 0.0  ; "when both are zero")]
    #[test_case(f64::INFINITY,  0.0, f64::INFINITY  ; "infinity")]
    // #[test_case(f64::NAN,  0.0, f64::NAN  ; "NaN")] // cant test NaN because NaN != NaN
    fn addition(a: f64, b: f64, e: f64) {
        let mut env = Environment::new();

        let a = Constant::new(a).into();
        let b = Constant::new(b).into();
        let d: NodeEnum = Add::new(a, b).into();

        let value = Rc::<NodeEnum>::try_unwrap(d.evaluate(&mut env).unwrap()).unwrap();

        assert!(
            value == Constant::new(e).into(),
            "Expected {} got {}",
            e,
            value
        )
    }

    #[test_case(69.0, 420.0, -351.0 ; "when both are positive")]
    #[test_case(-2.0,  -4.0, 2.0  ; "when both are negative")]
    #[test_case(0.0,  0.0, 0.0  ; "when both are zero")]
    #[test_case(f64::INFINITY,  0.0, f64::INFINITY  ; "infinity")]
    // #[test_case(f64::NAN,  0.0, f64::NAN  ; "NaN")] // cant test NaN because NaN != NaN
    fn subtraction(aa: f64, bb: f64, e: f64) {
        let mut env = Environment::new();

        let a = Constant::new(0.0).into();
        let b = Constant::new(aa).into();
        let c = Constant::new(bb).into();
        let d: NodeEnum = Subtract::new(Add::new(a, b).into(), c).into();

        let value = Rc::<NodeEnum>::try_unwrap(d.evaluate(&mut env).unwrap()).unwrap();

        assert!(
            value == Constant::new(e).into(),
            "Expected {} got {}",
            e,
            value
        )
    }

    #[test_case(5.0, 10.0, 50.0 ; "when both are positive")]
    #[test_case(-5.0, 10.0, -50.0 ; "when left is negative")]
    #[test_case(5.0, -10.0, -50.0 ; "when right is negative")]
    #[test_case(-5.0, -10.0, 50.0 ; "when both are negative")]
    #[test_case(2734589235234.23, 0.0, 0.0 ; "when 0 is involved")]
    fn multiplication(a: f64, b: f64, e: f64) {
        let mut env = Environment::new();

        let a = Constant::new(a).into();
        let b = Constant::new(b).into();
        let d: NodeEnum = Multiply::new(a, b).into();

        let value = Rc::<NodeEnum>::try_unwrap(d.evaluate(&mut env).unwrap()).unwrap();

        assert!(
            value == Constant::new(e).into(),
            "Expected {} got {}",
            e,
            value
        )
    }

    #[test_case(5.0, 10.0, 0.5 ; "when both are positive")]
    #[test_case(-5.0, 10.0, -0.5 ; "when left is negative")]
    #[test_case(5.0, -10.0, -0.5 ; "when right is negative")]
    #[test_case(-5.0, -10.0, 0.5 ; "when both are negative")]
    fn division(a: f64, b: f64, e: f64) {
        let mut env = Environment::new();

        let a = Constant::new(a).into();
        let b = Constant::new(b).into();
        let d: NodeEnum = Divide::new(a, b).into();

        let value = Rc::<NodeEnum>::try_unwrap(d.evaluate(&mut env).unwrap()).unwrap();

        assert!(
            value == Constant::new(e).into(),
            "Expected {} got {}",
            e,
            value
        )
    }
}

mod functions {
    #[test]
    fn stringification() {
        // let f = Function
    }
}

mod expected_errors {
    use test_case::test_case;

    #[test_case(5.0, 0.0 ; "divide by zero")]
    fn division(a: f64, b: f64) {
        let _ = a+b;
    }
}

mod misc {
    use test_case::test_case;

    #[test_case(30, '+', 60, '-', 20, "(30+60)-20" ; "add and subtract")]
    fn convert_to_string(
        a: impl Into<f64>,
        op1: char,
        b: impl Into<f64>,
        op2: char,
        c: impl Into<f64>,
        e: &'static str,
    ) {
    }
}
